@@ -3,9 +3,11 @@ import uuid
 import time
 import json
 from enum import Enum
-from tinydb import TinyDB, Query
+from tinydb import Query, where
 import concurrent
 from concurrent.futures.thread import ThreadPoolExecutor
+from operator import eq
+from functools import reduce
 
 from flask import request, jsonify
 from flask_restplus import Namespace, Resource
@@ -30,7 +32,7 @@ from portainer_api.models import AuthenticateUserRequest
 from portainer_api.rest import ApiException
 from .portainer_docker import PortainerDockerApi
 from .endpoints_api import EndpointsApi
-
+from .ansible_job import AnsibleJob
 
 class State(Enum):
     UNKNOWN = 0
@@ -45,6 +47,20 @@ class EPType(Enum):
     KUBERNETES = 2
     DOCKER = 3
 
+class QueryUser:
+    def query_builder(self, user=None, group=None, qargs=dict()):
+        qs = list()
+        if user:
+            qs.append(where('users').any(user) | eq(where('user'), user))
+        if group:
+            qs.append(where('groups').any(group))
+        for k,v in qargs.items():
+            if v:
+                qs.append(eq(where(k), v))
+        if len(qs):
+            return reduce(lambda a, b: a & b, qs)
+        return None
+
 # Basic auth
 httpauth = HTTPBasicAuth()
 
@@ -67,6 +83,12 @@ def verify_password(username, password):
             check_password_hash(users.get(username), password):
         return username
 
+def get_authinfo(request):
+    user = request.args.get('user', None)
+    group = request.args.get('group', None)
+    log.debug(f"User: {user}, Group: {group}")
+    return (user,group)
+
 class auth(object):
     def __init__(self, func):
         self.func = func
@@ -109,6 +131,7 @@ class auth(object):
         pcfg.api_key_prefix = {'Authorization': 'Bearer'}
 
         log.debug("Authenticating with token: {}".format(res.jwt))
+        pclient.jwt = res.jwt
         auth_expire = time.time() + 14400
 
         return pclient
@@ -116,30 +139,25 @@ class auth(object):
 
 @ns.route('/active')
 @ns.route('/active/<int:id>')
-@ns.route('/active/<user>')
-class ActiveCollection(Resource):
+class ActiveCollection(Resource, QueryUser):
 
     @httpauth.login_required
-    def get(self, id=None, user=None):
+    def get(self, id=None):
         """
-        Returns dictionary of active requests
+        Returns active sessions
         """
-        DB = TinyDB(cfg.get_dbpath())
-        Doc = Query()
-        table = DB.table('active')
-        docs = list()
-        if id:
-            doc = table.get(doc_id=id)
-            if doc:
-                docs = [doc]
-        elif user:
-            docs = table.search(Doc.user == user)
+        table = cfg.db.table('active')
+        (user,group) = get_authinfo(request)
+        query = self.query_builder(user, group, {"id": id})
+        if query and id:
+            res = table.get(query)
+            if not res:
+                return {"error": "Not found"}, 404
+            return res
+        elif query:
+            return table.search(query)
         else:
-            docs = table.all()
-        ret = []
-        for d in docs:
-            ret.append({d.doc_id: d})
-        return ret
+            return table.all()
 
     @ns.response(204, 'Allocation successfully deleted.')
     @ns.response(404, 'Not found.')
@@ -150,11 +168,11 @@ class ActiveCollection(Resource):
         """
         Deletes an active allocation (e.g. stops containers)
         """
-        DB = TinyDB(cfg.get_dbpath())
-        Node = Query()
-        nodes = DB.table('nodes')
-        table = DB.table('active')
-        doc = table.get(doc_id=id)
+        nodes = cfg.db.table('nodes')
+        table = cfg.db.table('active')
+        (user,group) = get_authinfo(request)
+        query = self.query_builder(user, group, {"id": id})
+        doc = table.get(query)
         if doc == None:
             return {"error": "Not found", "id": id}, 404
 
@@ -166,6 +184,7 @@ class ActiveCollection(Resource):
         with ThreadPoolExecutor(max_workers=8) as executor:
             for k, v in allocations.items():
                 try:
+                    Node = Query()
                     n = nodes.search(Node.name == k)[0]
                     if not (cfg.dryrun):
                         for alloc in v:
@@ -192,7 +211,7 @@ class ActiveCollection(Resource):
 @ns.route('/nodes')
 @ns.route('/nodes/<node>')
 @ns.route('/nodes/<int:id>')
-class NodeCollection(Resource):
+class NodeCollection(Resource, QueryUser):
 
     @httpauth.login_required
     @auth
@@ -200,19 +219,26 @@ class NodeCollection(Resource):
         """
         Returns list of existing nodes
         """
-        DB = TinyDB(cfg.get_dbpath())
-        DB.clear_cache()
+        cfg.db.clear_cache()
+        (user,group) = get_authinfo(request)
         refresh = request.args.get('refresh', None)
         if refresh and refresh.lower() == 'true':
             log.info("Refreshing endpoint DB...")
             global pclient
             init_db(pclient, refresh=True)
-        table = DB.table('nodes')
-        if node or id:
-            Node = Query()
-            nodes = table.search(Node.id == id) if id else table.search(Node.name == node)
-            return nodes if nodes else list()
-        return table.all()
+        else:
+            init_db(pclient, refresh=False)
+        table = cfg.db.table('nodes')
+        query = self.query_builder(user, group, {"id": id, "name": node})
+        if query and (id or node):
+            res = table.get(query)
+            if not res:
+                return {"error": "Not found"}, 404
+            return res
+        elif query:
+            return table.search(query)
+        else:
+            return table.all()
 
     @ns.response(204, 'Node successfully deleted.')
     @ns.response(404, 'Not found.')
@@ -224,17 +250,18 @@ class NodeCollection(Resource):
         """
         if not node and not id:
             return {"error": "Must specify node name or id"}, 400
-        DB = TinyDB(cfg.get_dbpath())
         Node = Query()
-        nodes = DB.table('nodes')
-        doc = nodes.get(Node.id == id) if id else nodes.get(Node.name == node)
+        nodes = cfg.db.table('nodes')
+        (user,group) = get_authinfo(request)
+        query = self.query_builder(user, group, {"id": id, "name": node})
+        doc = nodes.get(query)
         if doc == None:
             return {"error": "Not found"}, 404
         eapi = EndpointsApi(pclient)
         try:
             eapi.endpoint_delete(doc.get('id'))
         except Exception as e:
-            return {"error": f"Could not remove endpoint: {e}"}
+            log.info("Could not remove portainer endpoint, ignoring...")
         nodes.remove(doc_ids=[doc.doc_id])
         return None, 204
 
@@ -250,7 +277,6 @@ class NodeCollection(Resource):
         if type(req) is dict:
             req = [req]
         log.debug(req)
-
         eps = list()
         try:
             for r in req:
@@ -292,7 +318,7 @@ class NodeCollection(Resource):
 @ns.response(400, 'Bad Request')
 @ns.response(503, 'Service unavailable')
 @ns.route('/create')
-class Create(Resource):
+class Create(Resource, QueryUser):
 
     @httpauth.login_required
     @auth
@@ -300,7 +326,7 @@ class Create(Resource):
         """
         Handle the creation of a container service
         """
-        svcs = dict()
+        (user,group) = get_authinfo(request)
         req = request.get_json()
         if not req:
             raise BadRequest("Body is empty")
@@ -308,19 +334,64 @@ class Create(Resource):
             req = [req]
         log.debug(req)
 
+        cfg.db.clear_cache()
+        ntable = cfg.db.table('nodes')
+        ptable = cfg.db.table('profiles')
+        itable = cfg.db.table('images')
+
+        # Do auth and resource availability checks first
+        create = list()
+        for r in req:
+            instances = r.get("instances", None)
+            profile = r.get("profile", None)
+            image = r.get("image", None)
+            if instances == None or profile == None or image == None:
+                raise BadRequest("Missing fields in POST data")
+            # Profile
+            if not profile or profile == "default":
+                profile = settings.DEFAULT_PROFILE
+            query = self.query_builder(user, group, {"name": profile})
+            prof = cfg.get_profile(profile, user, group)
+            if not prof:
+                return {"error": f"Profile {profile} not found"}, 404
+            # Image
+            # By default try to pull the specified image name even if
+            # we don't know about it
+            if not user and not group:
+                img = image
+            else:
+                query = self.query_builder(user, group, {"name": image})
+                img = itable.get(query)
+                if not img:
+                    return {"error": f"Image {image} not found"}, 404
+                img = image
+            # Nodes
+            for ep in instances:
+                query = self.query_builder(user, group, {"name": ep})
+                node = ntable.get(query)
+                if not node:
+                    return {"error": f"Node {ep} not found"}, 404
+                create.append(
+                    {"node": node,
+                     "profile": prof,
+                     "image": img,
+                     "kwargs": r.get("kwargs", dict())
+                     }
+                )
+
+        svcs = dict()
         try:
             # keep a running set of addresses and ports allocated for this request
             addrs_v4 = set()
             addrs_v6 = set()
             cports = set()
             sports = set()
-            for r in req:
-                for s in r['instances']:
-                    kwargs = r.get("kwargs", dict())
-                    if s not in svcs:
-                        svcs[s] = list()
-                    svcs[s].append(create_service(s, r['image'], r['profile'], addrs_v4, addrs_v6,
-                                                  cports, sports, **kwargs))
+            for r in create:
+                s = r['node']['name']
+                if s not in svcs:
+                    svcs[s] = list()
+                svcs[s].append(create_service(r['node'], r['image'], r['profile'], addrs_v4, addrs_v6,
+                                              cports, sports, **r['kwargs']))
         except Exception as e:
             import traceback
             traceback.print_exc()
@@ -329,30 +400,52 @@ class Create(Resource):
 
         # setup simple accounting
         record = {'uuid': str(uuid.uuid4()),
-                  'user': httpauth.current_user(),
+                  'user': user if user else httpauth.current_user(),
                   'state': State.INITIALIZED.name,
                   'allocations': dict()}
 
         dapi = PortainerDockerApi(pclient)
         # get an ID from the DB
         Id = precommit_db()
+        errs = False
         for k, v in svcs.items():
             for s in svcs[k]:
                 # the portainer node this service will start on
                 n = s['node']
-                img = s['image']
                 if (cfg.dryrun):
                     ret = {'Id': str(uuid.uuid4())}
                 else:
+                    # Docker-specific v4 vs v6 image registry nonsense. Need to abstract this away.
+                    try:
+                        img = s['image']
+                        handle_image(n, img, dapi, s['pull_image'])
+                    except Exception as e:
+                        log.error("Could not pull image {} on node {}, {}: {}".format(img,
+                                                                                      n['name'],
+                                                                                      e.reason,
+                                                                                      e.body))
+                        errs = error_svc(s, e)
+                        try:
+                            v6img = f"registry.ipv6.docker.com/{s['image']}"
+                            handle_image(n, v6img, dapi, s['pull_image'])
+                            s['image'] = v6img
+                        except Exception as e:
+                            log.error("Could not pull image {} on node {}, {}: {}".format(v6img,
+                                                                                          n['name'],
+                                                                                          e.reason,
+                                                                                          e.body))
+                            errs = error_svc(s, e)
+                            continue
+                    # clear any errors if image resolved
+                    s['errors'] = list()
                     try:
-                        handle_image(n, img, dapi)
                         name = f"janus_{Id}" if Id else None
-                        ret = dapi.create_container(n['id'], img, name, **s['docker_kwargs'])
+                        ret = dapi.create_container(n['id'], s['image'], name, **s['docker_kwargs'])
                     except ApiException as e:
                         log.error("Could not create container on {}: {}: {}".format(n['name'],
                                                                                     e.reason,
                                                                                     e.body))
-                        error_svc(s, e)
+                        errs = error_svc(s, e)
                         continue
 
                 if not (cfg.dryrun):
@@ -365,26 +458,32 @@ class Create(Resource):
                         log.error("Could not connect network on {}: {}: {}".format(n['name'],
                                                                                    e.reason,
                                                                                    e.body))
-                        error_svc(s, e)
+                        errs = error_svc(s, e)
                         continue
 
                 s['container_id'] = ret['Id']
+                s['container_name'] = name
                 if n['name'] not in record['allocations']:
                     record['allocations'].update({n['name']: list()})
                 record['allocations'][n['name']].append(ret['Id'])
-
                 del s['node']
 
         # complete accounting
+        record['id'] = Id
         record['services'] = svcs
         record['request'] = req
-        return commit_db(record, Id)
+        if errs:
+            precommit_db(Id, delete=True)
+            return {Id: record}
+        else:
+            return commit_db(record, Id)
+
 
 @ns.response(200, 'OK')
 @ns.response(404, 'Not found')
 @ns.response(503, 'Service unavailable')
 @ns.route('/start/<int:id>')
-class Start(Resource):
+class Start(Resource, QueryUser):
 
     @httpauth.login_required
     @auth
@@ -392,12 +491,12 @@ class Start(Resource):
         """
         Handle the starting of container services
         """
-        DB = TinyDB(cfg.get_dbpath())
-        Srv = Query()
-        table = DB.table('active')
-        ntable = DB.table('nodes')
+        table = cfg.db.table('active')
+        ntable = cfg.db.table('nodes')
+        (user,group) = get_authinfo(request)
+        query = self.query_builder(user, group, {"id": id})
         if id:
-            svc = table.get(doc_id=id)
+            svc = table.get(query)
         if not svc:
             return {"error": "id not found"}, 404
 
@@ -421,8 +520,6 @@ class Start(Resource):
                 if not (cfg.dryrun):
                     try:
                         dapi.start_container(node['id'], c)
-
-                        log.info("s: {}".format(s["qos"]))
                         if s['qos']: # is not None and s['qos'].isinstance(dict)
                             qos = s["qos"]
                             qos["container"] = c
@@ -435,6 +532,32 @@ class Start(Resource):
                         error_svc(s, e)
                         error = True
                         continue
+
+                #
+                # Ansible job is requested if configured
+                # - Enviroment variabls must be set to access Ansible Tower server:
+                #   TOWER_HOST, TOWER_USERNAME, TOWER_PASSWORD, TOWER_SSL_VERIFY
+                # - It may take some time for the ansible job to finish or timeout (300 seconds)
+                #
+                prof = cfg.get_profile(s['profile'])
+                for psname in prof['settings']['post_starts']:
+                    ps = cfg.get_poststart(psname)
+                    if ps['type'] == 'ansible':
+                        jt_name = ps['jobtemplate']
+                        gateway = ps['gateway']
+                        ipprot = ps['ipprot']
+                        inf = ps['interface']
+                        limit = ps['limit']
+                        default_name= ps['container_name']
+                        container_name= s.get('container_name', default_name)
+                        ex_vars = f'{{"ipprot": "{ipprot}", "interface": "{inf}", "gateway": "{gateway}", "container": "{container_name}"}}'
+                        job = AnsibleJob()
+                        try:
+                            result = job.launch(job_template=jt_name, monitor=True, wait=True, timeout=600, extra_vars=ex_vars, limits=limit)
+                        except Exception as e:
+                            error_svc(s, e)
+                            continue
+                # End of Ansible job
         svc['state'] = State.MIXED.name if error else State.STARTED.name
         return commit_db(svc, id, realized=True)
 
@@ -442,7 +565,7 @@ class Start(Resource):
 @ns.response(404, 'Not found')
 @ns.response(503, 'Service unavailable')
 @ns.route('/stop/<int:id>')
-class Stop(Resource):
+class Stop(Resource, QueryUser):
 
     @httpauth.login_required
     @auth
@@ -450,10 +573,9 @@ class Stop(Resource):
         """
         Handle the stopping of container services
         """
-        DB = TinyDB(cfg.get_dbpath())
         Srv = Query()
-        table = DB.table('active')
-        ntable = DB.table('nodes')
+        table = cfg.db.table('active')
+        ntable = cfg.db.table('nodes')
         if id:
             svc = table.get(doc_id=id)
         if not svc:
@@ -501,6 +623,9 @@ class Exec(Resource):
         Handle the execution of a container command inside Service
         """
         svcs = dict()
+        start = False
+        attach = True
+        tty = False
         req = request.get_json()
         if type(req) is not dict or "Cmd" not in req:
             return {"error": "invalid request format"}, 400
@@ -513,10 +638,15 @@ class Exec(Resource):
         log.debug(req)
 
         nname = req["node"]
+        if "start" in req:
+            start = req["start"]
+        if "attach" in req:
+            attach = req["attach"]
+        if "tty" in req:
+            tty = req["tty"]
 
-        DB = TinyDB(cfg.get_dbpath())
         Node = Query()
-        table = DB.table('nodes')
+        table = cfg.db.table('nodes')
         node = table.get(Node.name == nname)
         if not node:
             return {"error": "Node not found: {}".format(nname)}
@@ -526,14 +656,15 @@ class Exec(Resource):
 
         dapi = PortainerDockerApi(pclient)
         kwargs = {'AttachStdin': False,
-                  'AttachStdout': True,
-                  'AttachStderr': True,
-                  'Tty': True,
+                  'AttachStdout': attach,
+                  'AttachStderr': attach,
+                  'Tty': tty,
                   'Cmd': cmd
                   }
         try:
             ret = dapi.exec_create(node["id"], container, **kwargs)
-            ret = dapi.exec_start(node["id"], ret["Id"])
+            if start:
+                ret = dapi.exec_start(node["id"], ret["Id"])
         except ApiException as e:
             log.error("Could not exec in container on {}: {}: {}".format(nname,
                                                                          e.reason,
@@ -541,11 +672,11 @@ class Exec(Resource):
             return {"error": e.reason}, 503
         return ret
 
-
 @ns.response(200, 'OK')
 @ns.response(503, 'Service unavailable')
 @ns.route('/qos')
 class QoS(Resource):
+
     @httpauth.login_required
     def get(self):
         name = request.args.get('name', None)
@@ -555,16 +686,38 @@ class QoS(Resource):
 
         return cfg.get_qos_list()
 
+@ns.response(200, 'OK')
+@ns.response(503, 'Service unavailable')
+@ns.route('/images')
+@ns.route('/images/<path:name>')
+class Images(Resource, QueryUser):
+
+    @httpauth.login_required
+    def get(self, name=None):
+        (user,group) = get_authinfo(request)
+        query = self.query_builder(user, group, {"name": name})
+        table = cfg.db.table('images')
+        if name:
+            res = table.get(query)
+            if not res:
+                return {"error": "Not found"}, 404
+            return res
+        elif query:
+            return table.search(query)
+        else:
+            return table.all()
 
 @ns.response(200, 'OK')
 @ns.response(503, 'Service unavailable')
 @ns.route('/profiles')
+@ns.route('/profiles/<name>')
 class Profile(Resource):
+
     @httpauth.login_required
-    def get(self):
+    def get(self, name=None):
         refresh = request.args.get('refresh', None)
         reset = request.args.get('reset', None)
-        pname = request.args.get('pname', None)
+        (user,group) = get_authinfo(request)
         if refresh and refresh.lower() == 'true':
             try:
                 cfg.read_profiles()
@@ -577,18 +730,18 @@ class Profile(Resource):
             except Exception as e:
                 return {"error": str(e)}, 500
 
-        if pname:
-            res = cfg.get_profile(pname, inline=True)
+        if name:
+            res = cfg.get_profile(name, user, group, inline=True)
             if not res:
-                return {"error": "Profile not found: {}".format(pname)}, 404
-
+                return {"error": "Profile not found: {}".format(name)}, 404
             return res
         else:
-            log.info("Returning all profiles")
-            return cfg.get_profiles(inline=True)
+            log.debug("Returning all profiles")
+            ret = cfg.get_profiles(user, group, inline=True)
+            return ret if ret else list()
 
     @httpauth.login_required
-    def post(self):
+    def post(self, name=None):
         try:
             req = request.get_json()
 
@@ -597,16 +750,15 @@ class Profile(Resource):
                 res.status_code = 400
                 return res
 
-            if "name" not in req or "settings" not in req:
-                res = jsonify(error="please follow this format: {\"name\": \"myprofile\", \"settings\": {\"key\": \"value\"}}")
+            if "settings" not in req:
+                res = jsonify(error="please follow this format: {\"settings\": {\"key\": \"value\"}}")
                 res.status_code = 400
                 return res
 
             configs = req["settings"]
-            pname = req["name"]
-            res = cfg.get_profile(pname, inline=True)
+            res = cfg.get_profile(name, inline=True)
             if res:
-                return {"error": "Profile {} already exists!".format(pname)}, 400
+                return {"error": "Profile {} already exists!".format(name)}, 400
 
             default = cfg._base_profile.copy()
             default.update((k, configs[k]) for k in default.keys() & configs.keys())
@@ -619,47 +771,42 @@ class Profile(Resource):
             return str(e), 500
 
         try:
-            DB = cfg._DB #TinyDB(cfg.get_dbpath())
-            profile_tbl = DB.table('profiles')
+            profile_tbl = cfg.db.table('profiles')
             log.info("Creating profile {}".format(
                 profile_tbl.insert({
-                'name': pname,
+                'name': name,
                 "settings": default
                 })
             ))
         except Exception as e:
             return str(e), 500
 
-        return cfg.get_profile(pname), 200
+        return cfg.get_profile(name), 200
 
     @httpauth.login_required
-    def put(self):
+    def put(self, name=None):
         try:
+            (user,group) = get_authinfo(request)
             req = request.get_json()
 
             if (req is None) or (req and type(req) is not dict):
                 res = jsonify(error="Body is not json dictionary")
-                res.status_code = 400
-                return res
+                raise BadRequest(res)
 
-            if "name" not in req or "settings" not in req:
-                res = jsonify(error="please follow this format: {\"name\": \"myprofile\", \"settings\": {\"key\": \"value\"}}")
-                res.status_code = 400
-                return res
+            if "settings" not in req:
+                res = jsonify(error="please follow this format: {\"settings\": {\"key\": \"value\"}}")
+                raise BadRequest(res)
 
             configs = req["settings"]
-            pname = req["name"]
-            pname = req["name"]
-            if pname == "default":
+            if name == "default":
                 return {"error": "Cannot update default profile!"}, 400
 
-            res = cfg.get_profile(pname, inline=True)
+            res = cfg.get_profile(name, user, group, inline=True)
             if not res:
-                return {"error": "Profile not found: {}".format(pname)}, 404
+                return {"error": "Profile not found: {}".format(name)}, 404
 
             default = res.copy()
-            log.info(default)
-            default.update((k, configs[k]) for k in default.keys() & configs.keys())
+            default['settings'].update(configs)
             ProfileSchema(**default)
 
         except ValidationError as e:
@@ -669,49 +816,126 @@ class Profile(Resource):
             return {"error" : str(e)}, 500
 
         try:
-            DB = cfg._DB #TinyDB(cfg.get_dbpath())
             query = Query()
-            profile_tbl = DB.table('profiles')
-            profile_tbl.upsert({
-                "settings": default
-            }, query.name == pname)
+            profile_tbl = cfg.db.table('profiles')
+            profile_tbl.update(default, query.name == name)
         except Exception as e:
             return str(e), 500
 
-        return cfg.get_profile(pname), 200
+        return cfg.get_profile(name), 200
 
     @httpauth.login_required
-    def delete(self):
+    def delete(self, name=None):
         try:
-            req = request.get_json()
+            (user,group) = get_authinfo(request)
 
-            if (req is None) or (req and type(req) is not dict):
-                res = jsonify(error="Body is not json dictionary")
-                res.status_code = 400
-                return res
+            if not name:
+                raise BadRequest("Must specify profile name")
 
-            if "name" not in req:
-                res = jsonify(error="please follow this format: {\"name\": \"myprofile\"}")
-                res.status_code = 400
-                return res
+            if name == "default":
+                raise BadRequest("Cannot delete default profile")
 
-            pname = req["name"]
-            if pname == "default":
-                return {"error": "Cannot delete default profile"}, 400
-
-            res = cfg.get_profile(pname, inline=True)
+            res = cfg.get_profile(name, user, group, inline=True)
             if not res:
-                return {"error": "Profile not found: {}".format(pname)}, 404
+                return {"error": "Profile not found: {}".format(name)}, 404
 
         except Exception as e:
             return str(e), 500
 
         try:
-            DB = cfg._DB #TinyDB(cfg.get_dbpath())
             query = Query()
-            profile_tbl = DB.table('profiles')
-            profile_tbl.remove(query.name == pname)
+            profile_tbl = cfg.db.table('profiles')
+            profile_tbl.remove(query.name == name)
         except Exception as e:
             return str(e), 500
 
         return {}, 204
+
+@ns.response(200, 'OK')
+@ns.response(204, 'Not modified')
+@ns.response(404, 'Not found')
+@ns.response(503, 'Service unavailable')
+@ns.route('/auth/<path:resource>')
+@ns.route('/auth/<path:resource>/<int:rid>')
+@ns.route('/auth/<path:resource>/<path:rname>')
+class JanusAuth(Resource):
+    resources = ["nodes",
+                 "images",
+                 "profiles",
+                 "active"]
+    get_resources = ["jwt"]
+
+    def _marshall_req(self):
+        req = request.get_json()
+        if not req:
+            raise BadRequest("Body is empty")
+        if type(req) is not dict:
+            raise BadRequest("Malformed data, expecting dict")
+        users = req.get("users", None)
+        groups = req.get("groups", None)
+        if users == None or groups == None:
+            raise BadRequest("users and groups not present in POST data")
+        log.debug(req)
+        return (users, groups)
+
+    def query_builder(self, id=None, name=None):
+        qs = list()
+        if id:
+            qs.append(eq(where('id'), id))
+        elif name:
+            qs.append(eq(where('name'), name))
+        return reduce(lambda a, b: a & b, qs)
+
+    @httpauth.login_required
+    @auth
+    def get(self, resource, rid=None, rname=None):
+        if resource not in self.resources and resource not in self.get_resources:
+            return {"error": f"Invalid resource path: {resource}"}, 404
+        # Returns active token for backend client (e.g. Portainer)
+        if resource in self.get_resources:
+            global plient
+            return {"jwt": pclient.jwt}, 200
+        query = self.query_builder(rid, rname)
+        table = cfg.db.table(resource)
+        res = table.get(query)
+        if not res:
+            return {"error": f"{resource} resource not found with id {rid if rid else rname}"}, 404
+        users = res.get("users", list())
+        groups = res.get("groups", list())
+        return {"users": users, "groups": groups}, 200
+
+    @httpauth.login_required
+    def post(self, resource, rid=None, rname=None):
+        (users, groups) = self._marshall_req()
+        query = self.query_builder(rid, rname)
+        table = cfg.db.table(resource)
+        res = table.get(query)
+        if not res:
+            return {"error": f"{resource} resource not found with id {rid if rid else rname}"}, 404
+        new_users = list(set(users).union(set(res.get("users", list()))))
+        new_groups = list(set(groups).union(set(res.get("groups", list()))))
+        res['users'] = new_users
+        res['groups'] = new_groups
+        table.update(res, query)
+        return res, 200
+
+    @httpauth.login_required
+    def delete(self, resource, rid=None, rname=None):
+        (users, groups) = self._marshall_req()
+        query = self.query_builder(rid, rname)
+        table = cfg.db.table(resource)
+        res = table.get(query)
+        if not res:
+            return {"error": f"{resource} resource not found with id {rid if rid else rname}"}, 404
+        for u in users:
+            try:
+                res['users'].remove(u)
+            except:
+                pass
+        for g in groups:
+            try:
+                res['groups'].remove(g)
+            except:
+                pass
+        table.update(res, query)
+        return res, 200
